# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker.api.validation import parameter_types

from tacker.sol_refactored.api.schemas import common_types


# SOL003 5.5.2.3
CreateVnfRequest_V200 = {
    'type': 'object',
    'properties': {
        'vnfdId': common_types.Identifier,
        'vnfInstanceName': {'type': 'string', 'maxLength': 255},
        'vnfInstanceDescription': {'type': 'string', 'maxLength': 1024},
        'metadata': parameter_types.keyvalue_pairs,
    },
    'required': ['vnfdId'],
    'additionalProperties': True,
}

# SOL003 5.5.2.4
InstantiateVnfRequest_V200 = {
    'type': 'object',
    'properties': {
        'flavourId': common_types.IdentifierInVnfd,
        'instantiationLevelId': common_types.IdentifierInVnfd,
        'extVirtualLinks': {
            'type': 'array',
            'items': common_types.ExtVirtualLinkData},
        'extManagedVirtualLinks': {
            'type': 'array',
            'items': common_types.ExtManagedVirtualLinkData},
        'vimConnectionInfo': {
            'type': 'object',
            'patternProperties': {
                '^.*$': common_types.VimConnectionInfo
            },
        },
        'localizationLanguage': {'type': 'string', 'maxLength': 255},
        'additionalParams': parameter_types.keyvalue_pairs,
        'extensions': parameter_types.keyvalue_pairs,
        'vnfConfigurableProperties': parameter_types.keyvalue_pairs
    },
    'required': ['flavourId'],
    'additionalProperties': True,
}

# SOL003 5.5.2.8
TerminateVnfRequest_V200 = {
    'type': 'object',
    'properties': {
        'terminationType': {
            'type': 'string',
            'enum': [
                'FORCEFUL',
                'GRACEFUL']
        },
        'gracefulTerminationTimeout': {
            'type': 'integer', 'minimum': 1
        },
        'additionalParams': parameter_types.keyvalue_pairs,
    },
    'required': ['terminationType'],
    'additionalProperties': True,
}

# SOL003 5.5.2.5
ScaleVnfRequest_V200 = {
    'type': 'object',
    'properties': {
        'type': {
            'type': 'string',
            'enum': ['SCALE_OUT', 'SCALE_IN']
        },
        'aspectId': common_types.IdentifierInVnfd,
        'numberOfSteps': {'type': 'integer', 'minimum': 1},
        'additionalParams': parameter_types.keyvalue_pairs,
    },
    'required': ['type', 'aspectId'],
    'additionalProperties': True,
}

# SOL002 5.5.2.9
# SOL003 5.5.2.9
HealVnfRequest_V200 = {
    'type': 'object',
    'properties': {
        'cause': {'type': 'string', 'maxLength': 255},
        'additionalParams': parameter_types.keyvalue_pairs,
        # NOTE: following fields support only NFV-SOL 002
        'vnfcInstanceId': {
            'type': 'array',
            'items': common_types.Identifier
        },
        'healScript': {'type': 'string', 'maxLength': 255},
    },
    'additionalProperties': True,
}

# SOL003 5.5.2.11
ChangeExtVnfConnectivityRequest_V200 = {
    'type': 'object',
    'properties': {
        'extVirtualLinks': {
            'type': 'array',
            'items': common_types.ExtVirtualLinkData
        },
        'vimConnectionInfo': {
            'type': 'object',
            'patternProperties': {
                '^.*$': common_types.VimConnectionInfo
            },
        },
        'additionalParams': parameter_types.keyvalue_pairs,
    },
    'required': ['extVirtualLinks'],
    'additionalProperties': True,
}

# SOL013 8.3.4
_SubscriptionAuthentication = {
    'type': 'object',
    'properties': {
        'authType': {
            'type': 'array',
            'items': {
                'type': 'string',
                'enum': [
                    'BASIC',
                    'OAUTH2_CLIENT_CREDENTIALS',
                    'TLS_CERT']
            }
        },
        'paramsBasic': {
            'type': 'object',
            'properties': {
                'userName': {'type': 'string'},
                'password': {'type': 'string'}
            },
            # NOTE: must be specified since the way to specify them out of
            # band is not supported.
            'required': ['userName', 'password']
        },
        'paramsOauth2ClientCredentials': {
            'type': 'object',
            'properties': {
                'clientId': {'type': 'string'},
                'clientPassword': {'type': 'string'},
                'tokenEndpoint': {'type': 'string'}
            },
            # NOTE: must be specified since the way to specify them out of
            # band is not supported.
            'required': ['clientId', 'clientPassword', 'tokenEndpoint']
        }
    },
    'required': ['authType'],
    'additionalProperties': True,
}

# SOL003 4.4.1.5 inner
_VnfProductVersions = {
    'type': 'array',
    'items': {
        'type': 'objects',
        'properties': {
            'vnfSoftwareVersion': {'type': 'string'},
            'vnfdVersions': {
                'type': 'array',
                'items': {'type': 'string'}
            }
        },
        'required': ['vnfSoftwareVersion'],
        'additionalProperties': True,
    }
}

# SOL003 4.4.1.5 inner
_VnfProducts = {
    'type': 'array',
    'items': {
        'type': 'object',
        'properties': {
            'vnfProductName': {'type': 'string'},
            'versions': _VnfProductVersions
        },
        'required': ['vnfProductName'],
        'additionalProperties': True,
    }
}

# SOL003 4.4.1.5 inner
_VnfProductsFromProviders = {
    'type': 'array',
    'items': {
        'type': 'object',
        'properties': {
            'type': 'object',
            'properties': {
                'vnfProvider': {'type': 'string'},
                'vnfProducts': _VnfProducts
            }
        },
        'required': ['vnfProvider'],
        'additionalProperties': True,
    }
}

# SOL003 4.4.1.5
_VnfInstanceSubscriptionFilter = {
    'type': 'object',
    'properties': {
        'vnfdIds': {
            'type': 'array',
            'items': common_types.Identifier
        },
        'vnfProductsFromProviders': _VnfProductsFromProviders,
        'vnfInstanceIds': {
            'type': 'array',
            'items': common_types.Identifier
        },
        'vnfInstanceNames': {
            'type': 'array',
            'items': {'type': 'string'}
        }
    },
    'additionalProperties': True,
}

# SOL003 5.5.3.12
_LifecycleChangeNotificationsFilter = {
    'type': 'object',
    'properties': {
        'vnfInstanceSubscriptionFilter': _VnfInstanceSubscriptionFilter,
        'notificationTypes': {
            'type': 'array',
            'items': {
                'type': 'string',
                'enum': [
                    'VnfLcmOperationOccurrenceNotification',
                    'VnfIdentifierCreationNotification',
                    'VnfIdentifierDeletionNotification']
            }
        },
        'operationTypes': {
            'type': 'array',
            'items': {
                'type': 'string',
                'enum': [
                    'INSTANTIATE',
                    'SCALE',
                    'SCALE_TO_LEVEL',
                    'CHANGE_FLAVOUR',
                    'TERMINATE',
                    'HEAL',
                    'OPERATE',
                    'CHANGE_EXT_CONN',
                    'MODIFY_INFO']
            }
        },
        'operationStates': {
            'type': 'array',
            'items': {
                'type': 'string',
                'enum': [
                    'STARTING',
                    'PROCESSING',
                    'COMPLETED',
                    'FAILED_TEMP',
                    'FAILED',
                    'ROLLING_BACK',
                    'ROLLED_BACK']
            }
        }
    },
    'additionalProperties': True,
}

# SOL003 5.5.2.15
LccnSubscriptionRequest_V200 = {
    'type': 'object',
    'properties': {
        'filter': _LifecycleChangeNotificationsFilter,
        'callbackUri': {'type': 'string', 'maxLength': 255},
        'authentication': _SubscriptionAuthentication,
        'verbosity': {
            'type': 'string',
            'enum': ['FULL', 'SHORT']
        }
    },
    'required': ['callbackUri'],
    'additionalProperties': True,
}

# SOL003 5.5.2.12
VnfInfoModificationRequest_V200 = {
    'type': 'object',
    'properties': {
        'vnfInstanceName': {'type': 'string', 'maxLength': 255},
        'vnfInstanceDescription': {'type': 'string', 'maxLength': 1024},
        'vnfdId': common_types.Identifier,
        'vnfConfigurableProperties': parameter_types.keyvalue_pairs,
        'metadata': parameter_types.keyvalue_pairs,
        'extensions': parameter_types.keyvalue_pairs,
        'vimConnectionInfo': {
            'type': 'object',
            'patternProperties': {
                '^.*$': common_types.VimConnectionInfo
            },
        },
        'vnfcInfoModifications': {
            'type': 'array',
            'items': common_types.VnfcInfoModifications
        },
    },
    'additionalProperties': True,
}



# Copyright 2011 VMware, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# If ../tacker/__init__.py exists, add ../ to Python search path, so that
# it will override what happens to be installed in /usr/(local/)lib/python...

import sys

from oslo_config import cfg
import oslo_i18n
from oslo_service import service as common_service

from tacker._i18n import _
from tacker.common import config
from tacker import objects
from tacker import service
from tacker.sol_refactored import objects as sol_objects


oslo_i18n.install("tacker")


def main():
    # the configuration will be read into the cfg.CONF global data structure
    config.init(sys.argv[1:])
    objects.register_all()
    sol_objects.register_all()
    if not cfg.CONF.config_file:
        sys.exit(_("ERROR: Unable to find configuration file via the default"
                   " search paths (~/.tacker/, ~/, /etc/tacker/, /etc/) and"
                   " the '--config-file' option!"))

    try:
        tacker_api = service.serve_wsgi(service.TackerApiService)
        launcher = common_service.launch(cfg.CONF, tacker_api,
                                         workers=cfg.CONF.api_workers or None,
                                         restart_method='mutate')
        launcher.wait()
    except KeyboardInterrupt:
        pass
    except RuntimeError as e:
        sys.exit(_("ERROR: %s") % e)



# Copyright 2013 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys
from unittest import mock

from tacker.tests import base
from tacker.tests import post_mortem_debug


class TestTesttoolsExceptionHandler(base.BaseTestCase):

    def test_exception_handler(self):
        try:
            self.assertTrue(False)
        except Exception:
            exc_info = sys.exc_info()
        with mock.patch('traceback.print_exception') as mock_print_exception:
            with mock.patch('pdb.post_mortem') as mock_post_mortem:
                with mock.patch.object(post_mortem_debug,
                                       'get_ignored_traceback',
                                       return_value=mock.Mock()):
                    post_mortem_debug.exception_handler(exc_info)

        # traceback will become post_mortem_debug.FilteredTraceback
        filtered_exc_info = (exc_info[0], exc_info[1], mock.ANY)
        mock_print_exception.assert_called_once_with(*filtered_exc_info)
        mock_post_mortem.assert_called_once_with(mock.ANY)


class TestFilteredTraceback(base.BaseTestCase):

    def test_filter_traceback(self):
        tb1 = mock.Mock()
        tb2 = mock.Mock()
        tb1.tb_next = tb2
        tb2.tb_next = None
        ftb1 = post_mortem_debug.FilteredTraceback(tb1, tb2)
        for attr in ['lasti', 'lineno', 'frame']:
            attr_name = 'tb_%s' % attr
            self.assertEqual(getattr(tb1, attr_name, None),
                             getattr(ftb1, attr_name, None))
        self.assertIsNone(ftb1.tb_next)


class TestGetIgnoredTraceback(base.BaseTestCase):

    def _test_get_ignored_traceback(self, ignored_bit_array, expected):
        root_tb = mock.Mock()

        tb = root_tb
        tracebacks = [tb]
        for x in range(len(ignored_bit_array) - 1):
            tb.tb_next = mock.Mock()
            tb = tb.tb_next
            tracebacks.append(tb)
        tb.tb_next = None

        tb = root_tb
        for ignored in ignored_bit_array:
            if ignored:
                tb.tb_frame.f_globals = ['__unittest']
            else:
                tb.tb_frame.f_globals = []
            tb = tb.tb_next

        actual = post_mortem_debug.get_ignored_traceback(root_tb)
        if expected is not None:
            expected = tracebacks[expected]
        self.assertEqual(expected, actual)

    def test_no_ignored_tracebacks(self):
        self._test_get_ignored_traceback([0, 0, 0], None)

    def test_single_member_trailing_chain(self):
        self._test_get_ignored_traceback([0, 0, 1], 2)

    def test_two_member_trailing_chain(self):
        self._test_get_ignored_traceback([0, 1, 1], 1)

    def test_first_traceback_ignored(self):
        self._test_get_ignored_traceback([1, 0, 0], None)

    def test_middle_traceback_ignored(self):
        self._test_get_ignored_traceback([0, 1, 0], None)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker.tests.functional.sol_kubernetes.vnflcm import base as vnflcm_base


class VnfLcmKubernetesMultiNsTest(vnflcm_base.BaseVnfLcmKubernetesTest):

    @classmethod
    def setUpClass(cls):
        super(VnfLcmKubernetesMultiNsTest, cls).setUpClass()
        vnf_package_id, cls.vnfd_id = cls._create_and_upload_vnf_package(
            cls, cls.tacker_client, "test_cnf_multi_ns",
            {"key": "sample_multi_ns_functional"})
        cls.vnf_package_ids.append(vnf_package_id)

    @classmethod
    def tearDownClass(cls):
        super(VnfLcmKubernetesMultiNsTest, cls).tearDownClass()

    def _test_cnf_scale(self, vnf_instance, aspect_id,
                        number_of_steps=1, error=False):
        scale_level = self._get_scale_level_by_aspect_id(
            vnf_instance, aspect_id)

        # test scale out
        scale_level = self._test_scale(
            vnf_instance['id'], 'SCALE_OUT', aspect_id, scale_level,
            number_of_steps, error)
        if error:
            return scale_level

        # test scale in
        scale_level = self._test_scale(
            vnf_instance['id'], 'SCALE_IN', aspect_id, scale_level,
            number_of_steps)

        return scale_level

    def test_multi_tenant_k8s_additional_params(self):
        vnf_instance_name = "multi_tenant_k8s_additional_params"
        vnf_instance_description = "multi tenant k8s additional params"
        files = ["Files/kubernetes/deployment_has_namespace.yaml",
                 "Files/kubernetes/namespace01.yaml"]
        additional_param = {
            "lcm-kubernetes-def-files": files,
            "namespace": "multi-namespace01"}
        # instantiate
        vnf_instance = self._create_and_instantiate_vnf_instance(
            self.vnfd_id, "simple", vnf_instance_name,
            vnf_instance_description, additional_param)
        # scale
        self._test_cnf_scale(vnf_instance, "vdu1_aspect", number_of_steps=1)

        before_vnfc_rscs = self._get_vnfc_resource_info(vnf_instance)
        deployment_target_vnfc = [vnfc_rsc for vnfc_rsc in before_vnfc_rscs if
                                  vnfc_rsc['vduId'] == 'VDU1'][0]
        vnfc_instance_id = [deployment_target_vnfc['id']]
        # heal
        after_vnfc_rscs = self._test_heal(vnf_instance, vnfc_instance_id)
        for vnfc_rsc in after_vnfc_rscs:
            after_pod_name = vnfc_rsc['computeResource']['resourceId']
            if vnfc_rsc['id'] == deployment_target_vnfc['id']:
                after_resource = deployment_target_vnfc
                compute_resource = after_resource['computeResource']
                before_pod_name = compute_resource['resourceId']
                self.assertNotEqual(after_pod_name, before_pod_name)
        # terminate
        self._terminate_vnf_instance(vnf_instance['id'])
        self._delete_vnf_instance(vnf_instance['id'])

    def test_multi_tenant_k8s_manifest(self):
        vnf_instance_name = "multi_tenant_k8s_manifest"
        vnf_instance_description = "multi tenant k8s manifest"
        files = ["Files/kubernetes/deployment_has_namespace.yaml",
                 "Files/kubernetes/namespace02.yaml"]
        additional_param = {"lcm-kubernetes-def-files": files}
        # instantiate
        vnf_instance = self._create_and_instantiate_vnf_instance(
            self.vnfd_id, "simple", vnf_instance_name,
            vnf_instance_description, additional_param)
        # scale
        self._test_cnf_scale(vnf_instance, "vdu1_aspect", number_of_steps=1)

        before_vnfc_rscs = self._get_vnfc_resource_info(vnf_instance)
        deployment_target_vnfc = [vnfc_rsc for vnfc_rsc in before_vnfc_rscs if
                                  vnfc_rsc['vduId'] == 'VDU1'][0]
        vnfc_instance_id = [deployment_target_vnfc['id']]
        # heal
        after_vnfc_rscs = self._test_heal(vnf_instance, vnfc_instance_id)
        for vnfc_rsc in after_vnfc_rscs:
            after_pod_name = vnfc_rsc['computeResource']['resourceId']
            if vnfc_rsc['id'] == deployment_target_vnfc['id']:
                after_resource = deployment_target_vnfc
                compute_resource = after_resource['computeResource']
                before_pod_name = compute_resource['resourceId']
                self.assertNotEqual(after_pod_name, before_pod_name)
        # terminate
        self._terminate_vnf_instance(vnf_instance['id'])
        self._delete_vnf_instance(vnf_instance['id'])

    def test_multi_tenant_k8s_default(self):
        vnf_instance_name = "multi_tenant_k8s_default"
        vnf_instance_description = "multi tenant k8s default"
        files = ["Files/kubernetes/deployment_no_namespace.yaml"]
        additional_param = {"lcm-kubernetes-def-files": files}
        # instantiate
        vnf_instance = self._create_and_instantiate_vnf_instance(
            self.vnfd_id, "simple", vnf_instance_name,
            vnf_instance_description, additional_param)
        # scale
        self._test_cnf_scale(vnf_instance, "vdu2_aspect", number_of_steps=1)

        before_vnfc_rscs = self._get_vnfc_resource_info(vnf_instance)
        deployment_target_vnfc = [vnfc_rsc for vnfc_rsc in before_vnfc_rscs if
                                  vnfc_rsc['vduId'] == 'VDU2'][0]
        vnfc_instance_id = [deployment_target_vnfc['id']]
        # heal
        after_vnfc_rscs = self._test_heal(vnf_instance, vnfc_instance_id)
        for vnfc_rsc in after_vnfc_rscs:
            after_pod_name = vnfc_rsc['computeResource']['resourceId']
            if vnfc_rsc['id'] == deployment_target_vnfc['id']:
                after_resource = deployment_target_vnfc
                compute_resource = after_resource['computeResource']
                before_pod_name = compute_resource['resourceId']
                self.assertNotEqual(after_pod_name, before_pod_name)
        # terminate
        self._terminate_vnf_instance(vnf_instance['id'])
        self._delete_vnf_instance(vnf_instance['id'])



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_db.sqlalchemy import utils as sqlalchemyutils

import tacker.context


def model_query(context, model, args=None, read_deleted=None,
                project_only=False):
    """Query helper that accounts for context's `read_deleted` field.

    :param context:     TackerContext of the query.
    :param model:       Model to query. Must be a subclass of ModelBase.
    :param args:        Arguments to query. If None - model is used.
    :param read_deleted: If not None, overrides context's read_deleted field.
                        Permitted values are 'no', which does not return
                        deleted values; 'only', which only returns deleted
                        values; and 'yes', which does not filter deleted
                        values.
    :param project_only: If set and context is user-type, then restrict
                        query to match the context's project_id. If set to
                        'allow_none', restriction includes project_id = None.
    """

    query_kwargs = {}
    if read_deleted:
        if 'no' == read_deleted:
            query_kwargs['deleted'] = False
        elif 'only' == read_deleted:
            query_kwargs['deleted'] = True
        elif 'yes' == read_deleted:
            pass
        else:
            raise ValueError(_("Unrecognized read_deleted value '%s'")
                             % read_deleted)

    query = sqlalchemyutils.model_query(
        model, context.session, args, **query_kwargs)

    if tacker.context.is_user_context(context) and project_only:
        query = query.filter_by(tenant_id=context.project_id)

    return query



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
from datetime import datetime as dt
import http.server
import inspect
import json
import os
import threading
import time
from urllib.parse import urlparse

from oslo_log import log as logging

LOG = logging.getLogger(__name__)


def PrepareRequestHandler(manager):
    class DummyRequestHandler(http.server.CGIHTTPRequestHandler):
        """HTTP request handler for dummy server."""

        def __init__(self, request, client_address, server):
            super().__init__(request, client_address, server)
            return

        def _is_match_with_list(self):
            """Return given path is listed in dictionary or not.

            Return:
                True/False
            """
            func_uri_list = manager._methods[self.command]
            for objChkUrl in func_uri_list:
                # Check which requested path is in our list.
                LOG.debug('path for check:%s' % objChkUrl)
                if(self.path.startswith(objChkUrl)):
                    return True
            return False

        def _returned_callback(self, path, mock_info):
            """Send responses to client. Called in do_* methods.

            This method do not handle message when error is occured.

            Args:
                path (str): URI path
                mock_info (tuple): callback informations from caller.
            """
            request_headers = dict(self.headers._headers)
            request_body = self._parse_request_body()
            response_body_str = b''

            (status_code, mock_headers, mock_body) = self._get_mock_info(
                mock_info, request_headers, request_body)
            self.send_response(status_code)

            # Check what I should return to client ?
            if mock_info.get('content') is not None:
                response_body_str = open(mock_info.get('content'), 'rb').read()
            elif len(mock_body) > 0:
                response_body_str = json.dumps(mock_body).encode('utf-8')
                mock_headers['Content-Length'] = str(len(response_body_str))

            # Send custom header if exist
            for key, val in mock_headers.items():
                self.send_header(key, val)
                self.end_headers()

            if len(response_body_str) > 0:
                self.wfile.write(response_body_str)

            manager.add_history(path, RequestHistory(
                status_code=status_code,
                request_headers=request_headers,
                request_body=request_body,
                response_headers=copy.deepcopy(mock_headers),
                response_body=copy.deepcopy(mock_body)))

            if mock_info.get('content') is None:
                self.end_headers()

        def _parse_request_body(self):
            if 'content-length' not in self.headers:
                return {}

            request_content_len = int(self.headers.get('content-length'))
            if request_content_len == 0:
                return {}

            decode_request_body = self.rfile.read(
                request_content_len).decode('utf-8')

            return json.loads(decode_request_body)

        def _get_mock_info(self, mock_info, request_headers, request_body):
            """Call mock(callback) and get responses

            This method is called from _returned_callback().

            Args:
                mock_info (tuple): callback informations from caller.
                request_headers (dict): Request headers
                request_body (dict):  Request Bodies

            Returns:
                (tuple): status_code, response headers, response bodies.
                         response body will be converted into JSON string
                         with json.dumps().
            """
            # Prepare response contents
            func = mock_info.get('callback')
            status_code = mock_info.get('status_code')
            mock_headers = mock_info.get('response_headers')
            mock_body = mock_info.get('response_body')

            # Call function if callable.
            if callable(func):
                mock_body = func(request_headers, request_body)

            return (status_code, mock_headers, mock_body)

        def do_DELETE(self):
            raise NotImplementedError

        def do_GET(self):
            """Process GET request"""
            LOG.debug('[Start] %s.%s()' % (self.__class__.__name__,
                inspect.currentframe().f_code.co_name))

            # Check URI in request.
            if self._is_match_with_list():
                # Request is registered in our list.
                tplUri = urlparse(self.path)
                self._returned_callback(tplUri.path,
                    manager._funcs_gets[tplUri.path])
            else:
                # Unregistered URI is requested
                LOG.debug('GET Recv. Unknown URL: "%s"' % self.path)
                self.send_response(http.HTTPStatus.BAD_REQUEST)
                self.end_headers()

            LOG.debug('[ End ] %s.%s()' %
                      (self.__class__.__name__,
                          inspect.currentframe().f_code.co_name))

        def do_POST(self):
            """Process POST request"""
            LOG.debug(
                '[Start] %s.%s()' %
                (self.__class__.__name__,
                 inspect.currentframe().f_code.co_name))

            # URI might have trailing uuid or not.
            if self._is_match_with_list():
                # Request is registered in our list.
                tplUri = urlparse(self.path)
                self._returned_callback(tplUri.path,
                    manager._funcs_posts[tplUri.path])
            else:
                # Unregistered URI is requested
                LOG.debug('POST Recv. Unknown URL: "%s"' % self.path)
                self.send_response(http.HTTPStatus.BAD_REQUEST)
                self.end_headers()

            LOG.debug(
                '[ End ] %s.%s()' %
                (self.__class__.__name__,
                 inspect.currentframe().f_code.co_name))

        def do_PUT(self):
            raise NotImplementedError

    return DummyRequestHandler


class RequestHistory:
    """Storage class for storing requested data(Maybe POSTed datas)."""

    def __init__(
            self,
            status_code,
            request_headers=None,
            request_body=None,
            response_headers=None,
            response_body=None):
        self.timestamp = dt.now()
        self.status_code = status_code
        self.request_headers = request_headers
        self.request_body = request_body
        self.response_headers = response_headers
        self.response_body = response_body


class FakeServerManager(object):
    """Manager class to manage dummy server setting and control"""

    SERVER_PORT = 9990
    SERVER_PORT_T1 = 9995
    SERVER_PORT_T2 = 9996
    SERVER_INVOKE_CHECK_INTERVAL = 10

    def __init__(self):
        # Initialize class-specific variables.

        # Storage for request header/body and response header/body
        # history (dict) is updated using RequestHistory class.
        self._history = {}

        # Initialize function list for each request method.
        # DELETE/PUT method is listed but not supported currently.
        self._funcs_deletes = {}
        self._funcs_gets = {}
        self._funcs_posts = {}
        self._funcs_puts = {}
        self._methods = {
            'DELETE': self._funcs_deletes,
            'GET': self._funcs_gets,
            'POST': self._funcs_posts,
            'PUT': self._funcs_puts}

    def set_callback(
            self,
            method,
            uri,
            status_code=None,
            response_headers=None,
            response_body=None,
            content=None,
            callback=None):
        """Set callback function and some stuff for specified URI.

        ALL additional parameter is set default to None, so you have to
        specify what your callback-function need. response_header and
        response_body will be passed to callback.

        Args:
            method (str): Reqested method
            uri (str): Requested URI
            status_code (http.HTTPStatus): HTTP status code and
                                           reason phrase.
            response_headers (dict): Addtional response header.
            response_body (dict): Response body. Must be Jason Bourne
            content (str): File path that you want client to download.
            callback (callable): Callback function. Must return serializable
                                 object json.dumps() can handle.
        """
        callbacks = self._methods[method]
        callbacks[uri] = {
            'status_code': status_code or http.HTTPStatus.OK,
            'response_headers': response_headers or {},
            'response_body': response_body or {},
            'content': content,
            'callback': callback,
        }
        self._methods[method].update(callbacks)

        # Check file existence for content
        if content is not None:
            if not os.path.isfile(content):
                raise FileNotFoundError

        LOG.debug('Set callback for %s(%s): %s' %
                  (method, uri, callback))

    def add_history(self, path, history):
        """Add Request/Response header/body to history.

        This method maybe called in DummyRequestHandler._returned_callback()
        only. This method should not be called from outside of This class.

        Args:
            path (str): URI path
            history (RequestHistory): Storage container for each request.
        """
        if path in self._history:
            self._history[path].append(history)
        else:
            self._history[path] = [history]

    def clear_history(self, path=None):
        """Clear Request/Response header/body of history.

        Args:
            path (str): URI path
        """
        if not path:
            self._history = {}
            return

        if path in self._history:
            self._history.pop(path)

    def get_history(self, path=None):
        """Get Request/Response header/body from history.

        Args:
            path (str): URI path

        Returns:
            history list(RequestHistory): Storage container for each request.
        """
        history = copy.deepcopy(self._history)

        if not path:
            return history

        return history.get(path) or []

    def prepare_http_server(
            self,
            address="localhost",
            port=SERVER_PORT):
        """Set up HTTPd server your behalf.

        Args:
            address (str): bind address for listen
            port (int): por number for listen
        """
        LOG.debug(
            '[Start] %s.%s()' %
            (self.__class__.__name__,
             inspect.currentframe().f_code.co_name))
        while True:
            try:
                RequestHandler = PrepareRequestHandler(self)
                self.objHttpd = http.server.HTTPServer(
                    (address, port), RequestHandler)
            except OSError:
                time.sleep(self.SERVER_INVOKE_CHECK_INTERVAL)
                continue
            else:
                break
        self.thread_server = threading.Thread(None, self.run)
        LOG.debug(
            '[ End ] %s.%s()' %
            (self.__class__.__name__,
             inspect.currentframe().f_code.co_name))

    def start_server(self):
        """Start server in thread."""
        LOG.debug('[START] %s()' % inspect.currentframe().f_code.co_name)
        self.thread_server.start()
        LOG.debug('[ END ] %s()' % inspect.currentframe().f_code.co_name)

    def run(self):
        """HTTPd server runner"""
        LOG.debug('[START] %s()' % inspect.currentframe().f_code.co_name)
        try:
            self.objHttpd.serve_forever()
        except KeyboardInterrupt:
            self.stop_server()
        finally:
            self.objHttpd.server_close()
        LOG.debug('[ END ] %s()' % inspect.currentframe().f_code.co_name)

    def stop_server(self):
        """Stop HTTP Server"""
        LOG.debug('[START] %s()' % inspect.currentframe().f_code.co_name)
        thread_shutdown = threading.Thread(None, self.objHttpd.shutdown)
        thread_shutdown.daemon = True
        thread_shutdown.start()
        self.thread_server.join()
        LOG.debug('[ END ] %s()' % inspect.currentframe().f_code.co_name)



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import importlib
import os
import pickle
import sys
import traceback


def main():
    script_dict = pickle.load(sys.stdin.buffer)

    req = script_dict['request']
    inst = script_dict['vnf_instance']
    grant_req = script_dict['grant_request']
    grant = script_dict['grant_response']
    tmp_csar_dir = script_dict['tmp_csar_dir']

    additional_params = req['additionalParams']
    userdata_path = additional_params['lcm-operation-user-data']
    userdata_class = additional_params['lcm-operation-user-data-class']

    sys.path.append(tmp_csar_dir)
    class_module = os.path.splitext(
        userdata_path.lstrip('./'))[0].replace('/', '.')
    module = importlib.import_module(class_module)
    klass = getattr(module, userdata_class)

    method = getattr(klass, grant_req['operation'].lower())
    stack_dict = method(req, inst, grant_req, grant, tmp_csar_dir)

    pickle.dump(stack_dict, sys.stdout.buffer)
    sys.stdout.flush()


if __name__ == "__main__":
    try:
        main()
        os._exit(0)
    except Exception:
        sys.stderr.write(traceback.format_exc())
        sys.stderr.flush()
        os._exit(1)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils

from tacker.common import utils
from tacker.objects import base
from tacker.objects import fields


@base.TackerObjectRegistry.register
class GrantRequest(base.TackerObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'vnf_instance_id': fields.StringField(nullable=False),
        'vnf_lcm_op_occ_id': fields.StringField(nullable=False),
        'vnfd_id': fields.StringField(nullable=False),
        'flavour_id': fields.StringField(nullable=True),
        'operation': fields.StringField(nullable=False),
        'is_automatic_invocation': fields.BooleanField(nullable=False,
                                                       default=False),
        'add_resources': fields.ListOfObjectsField(
            'ResourceDefinition', nullable=True, default=[]),
        'remove_resources': fields.ListOfObjectsField(
            'ResourceDefinition', nullable=True, default=[]),
        'update_resources': fields.ListOfObjectsField(
            'ResourceDefinition', nullable=True, default=[]),
        'placement_constraints': fields.ListOfObjectsField(
            'PlacementConstraint', nullable=True, default=[]),
        '_links': fields.ObjectField(
            'Links', nullable=False)
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_grant_req = super(
                GrantRequest, cls).obj_from_primitive(primitive, context)
        else:
            if 'add_resources' in primitive.keys():
                obj_data = [ResourceDefinition._from_dict(
                    add_rsc) for add_rsc in primitive.get(
                    'add_resources', [])]
                primitive.update({'add_resources': obj_data})
            if 'remove_resources' in primitive.keys():
                obj_data = [ResourceDefinition._from_dict(
                    remove_rsc) for remove_rsc in primitive.get(
                    'remove_resources', [])]
                primitive.update({'add_resources': obj_data})
            if 'update_resources' in primitive.keys():
                obj_data = [ResourceDefinition._from_dict(
                    update_rsc) for update_rsc in primitive.get(
                    'update_resources', [])]
                primitive.update({'update_resources': obj_data})
            if 'placement_constraints' in primitive.keys():
                obj_data = [PlacementConstraint._from_dict(
                    place) for place in primitive.get(
                    'placement_constraints', [])]
                primitive.update({'add_resources': obj_data})
            if '_links' in primitive.keys():
                obj_data = Links._from_dict(
                    primitive.get('_links', {}))
                primitive.update({'_links': obj_data})
            obj_grant_req = GrantRequest._from_dict(primitive)

        return obj_grant_req

    @classmethod
    def _from_dict(cls, data_dict):
        vnf_instance_id = data_dict.get('vnf_instance_id')
        vnf_lcm_op_occ_id = data_dict.get('vnf_lcm_op_occ_id')
        vnfd_id = data_dict.get('vnfd_id')
        flavour_id = data_dict.get('flavour_id')
        operation = data_dict.get('operation')
        is_automatic_invocation = data_dict.get('is_automatic_invocation')
        add_resources = data_dict.get('add_resources', [])
        remove_resources = data_dict.get('remove_resources', [])
        update_resources = data_dict.get('update_resources', [])
        placement_constraints = data_dict.get('placement_constraints', [])
        links = data_dict.get('_links')

        obj = cls(
            vnf_instance_id=vnf_instance_id,
            vnf_lcm_op_occ_id=vnf_lcm_op_occ_id,
            vnfd_id=vnfd_id,
            flavour_id=flavour_id,
            operation=operation,
            is_automatic_invocation=is_automatic_invocation,
            add_resources=add_resources,
            remove_resources=remove_resources,
            update_resources=update_resources,
            placement_constraints=placement_constraints,
            _links=links)
        return obj

    def to_dict(self):
        data = {'vnf_instance_id': self.vnf_instance_id,
                'vnf_lcm_op_occ_id': self.vnf_lcm_op_occ_id,
                'vnfd_id': self.vnfd_id,
                'flavour_id': self.flavour_id,
                'operation': self.operation,
                'is_automatic_invocation': self.is_automatic_invocation,
                '_links': self._links.to_dict()}
        if self.add_resources:
            add_resources_list = []
            for add_resource in self.add_resources:
                add_resources_list.append(add_resource.to_dict())

            data.update({'add_resources': add_resources_list})
        if self.remove_resources:
            remove_resources_list = []
            for remove_resource in self.remove_resources:
                remove_resources_list.append(remove_resource.to_dict())

            data.update({'remove_resources': remove_resources_list})
        if self.update_resources:
            update_resources_list = []
            for update_resource in self.update_resources:
                update_resources_list.append(update_resource.to_dict())

            data.update({'update_resources': update_resources_list})
        if self.placement_constraints:
            placement_constraints_list = []
            for placement_constraint in self.placement_constraints:
                placement_constraints_list.append(
                    placement_constraint.to_dict())

            data.update({'placement_constraints': placement_constraints_list})
        return data

    def to_request_body(self):
        req_dict = self.to_dict()
        req_dict = utils.convert_snakecase_to_camelcase(req_dict)
        return jsonutils.dumps(req_dict).replace('Links', '_links')


@base.TackerObjectRegistry.register
class ResourceDefinition(base.TackerObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.StringField(nullable=False),
        'type': fields.StringField(nullable=False),
        'vdu_id': fields.StringField(nullable=True, default=None),
        'resource_template_id': fields.StringField(nullable=False),
        'resource': fields.ObjectField(
            'ResourceHandle', nullable=True, default=None)
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_grant_req = super(
                ResourceDefinition, cls).\
                obj_from_primitive(primitive, context)
        else:
            if 'resource' in primitive.keys():
                obj_data = ResourceHandle._from_dict(
                    primitive.get('resource'))
                primitive.update({'resource': obj_data})
            obj_grant_req = ResourceDefinition._from_dict(primitive)

        return obj_grant_req

    @classmethod
    def _from_dict(cls, data_dict):
        id = data_dict.get('id')
        type = data_dict.get('type')
        vdu_id = data_dict.get('vdu_id')
        resource_template_id = data_dict.get('resource_template_id')
        resource = data_dict.get('resource')

        obj = cls(
            id=id,
            type=type,
            vdu_id=vdu_id,
            resource_template_id=resource_template_id,
            resource=resource)
        return obj

    def to_dict(self):
        data = {'id': self.id,
                'type': self.type,
                'resource_template_id': self.resource_template_id}
        if self.vdu_id:
            data.update({'vdu_id': self.vdu_id})
        if self.resource:
            data.update({'resource': self.resource.to_dict()})
        return data


@base.TackerObjectRegistry.register
class PlacementConstraint(base.TackerObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'affinity_or_anti_affinity': fields.StringField(nullable=False),
        'scope': fields.StringField(nullable=False),
        'resource': fields.ListOfObjectsField(
            'ConstraintResourceRef', nullable=False, default=[]),
        'fallback_best_effort': fields.BooleanField(nullable=False),
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_placement_constraint = super(
                PlacementConstraint, cls).obj_from_primitive(
                primitive, context)
        else:
            if 'resource' in primitive.keys():
                obj_data = [ConstraintResourceRef._from_dict(
                    add_rsc) for add_rsc in primitive.get(
                    'resource', [])]
                primitive.update({'resource': obj_data})
            obj_placement_constraint = PlacementConstraint._from_dict(
                primitive)

        return obj_placement_constraint

    @classmethod
    def _from_dict(cls, data_dict):
        affinity_or_anti_affinity = data_dict.get('affinity_or_anti_affinity')
        scope = data_dict.get('scope')
        resource = data_dict.get('resource')
        fallback_best_effort = data_dict.get('fallback_best_effort')

        obj = cls(
            affinity_or_anti_affinity=affinity_or_anti_affinity,
            scope=scope,
            resource=resource,
            fallback_best_effort=fallback_best_effort)
        return obj

    def to_dict(self):
        data = {'affinity_or_anti_affinity': self.affinity_or_anti_affinity,
                'scope': self.scope,
                'fallback_best_effort': self.fallback_best_effort}
        if self.resource:
            resource_list = []
            for rsc in self.resource:
                resource_list.append(rsc.to_dict())

            data.update({'resource': resource_list})
        return data


@base.TackerObjectRegistry.register
class ConstraintResourceRef(base.TackerObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id_type': fields.StringField(nullable=False),
        'resource_id': fields.StringField(nullable=False),
        'vim_connection_id': fields.StringField(nullable=True, default=None),
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_placement_constraint = super(
                ConstraintResourceRef, cls).obj_from_primitive(
                primitive, context)
        else:
            obj_placement_constraint = ConstraintResourceRef._from_dict(
                primitive)

        return obj_placement_constraint

    @classmethod
    def _from_dict(cls, data_dict):
        id_type = data_dict.get('id_type')
        resource_id = data_dict.get('resource_id')
        vim_connection_id = data_dict.get('vim_connection_id')

        obj = cls(
            id_type=id_type,
            resource_id=resource_id,
            vim_connection_id=vim_connection_id)
        return obj

    def to_dict(self):
        data = {'id_type': self.id_type,
                'resource_id': self.resource_id}
        if self.vim_connection_id:
            data.update({'vim_connection_id': self.vim_connection_id})
        return data


@base.TackerObjectRegistry.register
class Links(base.TackerObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'vnf_lcm_op_occ': fields.ObjectField(
            'Link', nullable=False),
        'vnf_instance': fields.ObjectField(
            'Link', nullable=False)
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_links = super(
                Links, cls).obj_from_primitive(primitive, context)
        else:
            if 'vnf_lcm_op_occ' in primitive.keys():
                obj_data = Link._from_dict(
                    primitive.get('vnf_lcm_op_occ'))
                primitive.update({'vnf_lcm_op_occ': obj_data})
            if 'vnf_instance' in primitive.keys():
                obj_data = Link._from_dict(
                    primitive.get('vnf_instance'))
                primitive.update({'vnf_instance': obj_data})
            obj_links = Links._from_dict(primitive)

        return obj_links

    @classmethod
    def _from_dict(cls, data_dict):
        vnf_lcm_op_occ = data_dict.get('vnf_lcm_op_occ')
        vnf_instance = data_dict.get('vnf_instance')

        obj = cls(
            vnf_lcm_op_occ=vnf_lcm_op_occ,
            vnf_instance=vnf_instance)
        return obj

    def to_dict(self):
        return {'vnf_lcm_op_occ': self.vnf_lcm_op_occ.to_dict(),
                'vnf_instance': self.vnf_instance.to_dict()}


@base.TackerObjectRegistry.register
class Link(base.TackerObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'href': fields.StringField(nullable=False)
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_link = super(
                Link, cls).obj_from_primitive(primitive, context)
        else:
            obj_link = Link._from_dict(primitive)

        return obj_link

    @classmethod
    def _from_dict(cls, data_dict):
        href = data_dict.get('href')

        obj = cls(
            href=href)
        return obj

    def to_dict(self):
        return {'href': self.href}


@base.TackerObjectRegistry.register
class ResourceHandle(base.TackerObject,
                     base.TackerPersistentObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'vim_connection_id': fields.StringField(nullable=True,
                                                default=None),
        'resource_id': fields.StringField(nullable=False, default=""),
        'vim_level_resource_type': fields.StringField(nullable=True,
                                                      default=None)
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            resource_handle = super(
                ResourceHandle, cls).obj_from_primitive(
                primitive, context)
        else:
            resource_handle = ResourceHandle._from_dict(primitive)

        return resource_handle

    @classmethod
    def _from_dict(cls, data_dict):
        vim_connection_id = data_dict.get('vim_connection_id')
        resource_id = data_dict.get('resource_id', "")
        vim_level_resource_type = data_dict.get('vim_level_resource_type')

        obj = cls(vim_connection_id=vim_connection_id,
                  resource_id=resource_id,
                  vim_level_resource_type=vim_level_resource_type)

        return obj

    def to_dict(self):
        return {'vim_connection_id': self.vim_connection_id,
                'resource_id': self.resource_id,
                'vim_level_resource_type': self.vim_level_resource_type}



# Copyright 2016 Brocade Communications Systems Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from tacker._i18n import _
from tacker.api import extensions
from tacker.api.v1 import attributes as attr
from tacker.api.v1 import resource_helper
from tacker.common import exceptions
from tacker.plugins.common import constants
from tacker.services import service_base


class VimUnauthorizedException(exceptions.TackerException):
    message = _("%(message)s")


class VimConnectionException(exceptions.TackerException):
    message = _("%(message)s")


class VimInUseException(exceptions.TackerException):
    message = _("VIM %(vim_id)s is still in use by VNF")


class VimDefaultNotDefined(exceptions.TackerException):
    message = _("Default VIM is not defined.")


class VimDefaultDuplicateException(exceptions.TackerException):
    message = _("Default VIM already exists %(vim_id)s.")


class VimNotFoundException(exceptions.TackerException):
    message = _("Specified VIM id %(vim_id)s is invalid. Please verify and "
                "pass a valid VIM id")


class VimRegionNotFoundException(exceptions.TackerException):
    message = _("Unknown VIM region name %(region_name)s")


class VimKeyNotFoundException(exceptions.TackerException):
    message = _("Unable to find key file for VIM %(vim_id)s")


class VimEncryptKeyError(exceptions.TackerException):
    message = _("Barbican must be enabled for VIM %(vim_id)s")


class VimUnsupportedResourceTypeException(exceptions.TackerException):
    message = _("Resource type %(type)s is unsupported by VIM")


class VimGetResourceException(exceptions.TackerException):
    message = _("Error while trying to issue %(cmd)s to find resource type "
                "%(type)s by resource name %(name)s")


class VimGetResourceNameNotUnique(exceptions.TackerException):
    message = _("Getting resource id from VIM with resource name %(name)s "
                "by %(cmd)s returns more than one")


class VimGetResourceNotFoundException(exceptions.TackerException):
    message = _("Getting resource id from VIM with resource name %(name)s "
                "by %(cmd)s returns nothing")


class VimFromVnfNotFoundException(exceptions.NotFound):
    message = _('VIM from VNF %(vnf_id)s could not be found')


class ToscaParserFailed(exceptions.InvalidInput):
    message = _("tosca-parser failed: - %(error_msg_details)s")


class VnffgdInvalidTemplate(exceptions.InvalidInput):
    message = _("Invalid VNFFG template input: %(template)s")


class VnffgdDuplicateForwarderException(exceptions.InvalidInput):
    message = _("Invalid Forwarding Path contains duplicate forwarder not in "
                "order: %(forwarder)s")


class VnffgdDuplicateCPException(exceptions.InvalidInput):
    message = _("Invalid Forwarding Path contains duplicate connection point "
                ": %(cp)s")


class VnffgdVnfdNotFoundException(exceptions.NotFound):
    message = _("Specified VNFD %(vnfd_name)s in VNFFGD does not exist. "
                "Please create VNFDs before creating VNFFG")


class VnffgdCpNotFoundException(exceptions.NotFound):
    message = _("Specified CP %(cp_id)s could not be found in VNFD "
                "%(vnfd_name)s. Please check VNFD for correct Connection "
                "Point.")


class VnffgdCpNoForwardingException(exceptions.TackerException):
    message = _("Specified CP %(cp_id)s in VNFD %(vnfd_name)s "
                "does not have forwarding capability, which is required to be "
                "included in forwarding path")


class VnffgdWrongEndpointNumber(exceptions.TackerException):
    message = _("Specified number_of_endpoints %(number)s is not equal to "
                "the number of connection_point %(cps)s")


class VnffgdInUse(exceptions.InUse):
    message = _('VNFFGD %(vnffgd_id)s is still in use')


class VnffgdNotFoundException(exceptions.NotFound):
    message = _('VNFFG Template %(vnffgd_id)s could not be found')


class VnffgCreateFailed(exceptions.TackerException):
    message = _('Creating VNFFG based on %(vnffgd_id)s failed')


class VnffgInvalidMappingException(exceptions.TackerException):
    message = _("Matching VNF Instance for VNFD %(vnfd_name)s could not be "
                "found. Please create an instance of this VNFD before "
                "creating/updating VNFFG.")


class VnffgParamValueFormatError(exceptions.TackerException):
    message = _("Param values %(param_value)s is not in dict format.")


class VnffgTemplateParamParsingException(exceptions.TackerException):
    message = _("Failed to parse VNFFG Template due to "
                "missing input param %(get_input)s.")


class VnffgPropertyNotFoundException(exceptions.NotFound):
    message = _('VNFFG Property %(vnffg_property)s could not be found')


class VnffgCpNotFoundException(exceptions.NotFound):
    message = _("Specified CP %(cp_id)s could not be found in VNF "
                "%(vnf_id)s.")


class VnffgNotFoundException(exceptions.NotFound):
    message = _('VNFFG %(vnffg_id)s could not be found')


class VnffgInUse(exceptions.InUse):
    message = _('VNFFG %(vnffg_id)s is still in use')


class VnffgVnfNotFoundException(exceptions.NotFound):
    message = _("Specified VNF instance %(vnf_name)s in VNF Mapping could not "
                "be found")


class VnffgDeleteFailed(exceptions.TackerException):
    message = _('Deleting VNFFG %(vnffg_id)s failed')


class VnffgInUseNS(exceptions.TackerException):
    message = _('VNFFG %(vnffg_id)s belongs to active network service '
                '%(ns_id)s')


class NfpAttributeNotFoundException(exceptions.NotFound):
    message = _('NFP attribute %(attribute)s could not be found')


class NfpNotFoundException(exceptions.NotFound):
    message = _('NFP %(nfp_id)s could not be found')


class NfpInUse(exceptions.InUse):
    message = _('NFP %(nfp_id)s is still in use')


class NfpPolicyCriteriaError(exceptions.PolicyCheckError):
    message = _('%(error)s in policy')


class NfpPolicyCriteriaIndexError(exceptions.TackerException):
    message = _('Criteria list can not be empty')


class NfpDuplicatePolicyCriteria(exceptions.TackerException):
    message = _('The %(first_dict)s and %(sec_dict)s are overlapped')


class NfpDuplicatePathID(exceptions.TackerException):
    message = _('The path_id %(path_id)s is overlapped with '
                'NFP %(nfp_name)s in %(vnffg_name)s')


class NfpPolicyTypeError(exceptions.PolicyCheckError):
    message = _('Unsupported Policy Type: %(type)s')


class NfpForwarderNotFoundException(exceptions.NotFound):
    message = _('VNFD Forwarder %(vnfd)s not found in VNF Mapping %(mapping)s')


class NfpRequirementsException(exceptions.TackerException):
    message = _('VNFD Forwarder %(vnfd)s specified more than twice in '
                'requirements path')


class SfcInUse(exceptions.InUse):
    message = _('SFC %(sfc_id)s is still in use')


class SfcNotFoundException(exceptions.NotFound):
    message = _('Service Function Chain %(sfc_id)s could not be found')


class ClassifierInUse(exceptions.InUse):
    message = _('Classifier %(classifier_id)s is still in use')


class ClassifierNotFoundException(exceptions.NotFound):
    message = _('Classifier %(classifier_id)s could not be found')


class VnfMappingNotFoundException(exceptions.NotFound):
    message = _('VNF mapping not found/defined')


class VnfMappingNotValidException(exceptions.TackerException):
    message = _('The %(vnfd)s is not found in constituent VNFDs')


class NSDInUse(exceptions.InUse):
    message = _('NSD %(nsd_id)s is still in use')


class NSInUse(exceptions.InUse):
    message = _('NS %(ns_id)s is still in use')


class NoTasksException(exceptions.TackerException):
    message = _('No tasks to run for %(action)s on %(resource)s')


class UpdateChainException(exceptions.TackerException):
    message = _("%(message)s")


class CreateChainException(exceptions.TackerException):
    message = _("%(message)s")


class UpdateClassifierException(exceptions.TackerException):
    message = _("%(message)s")


class UpdateVnffgException(exceptions.TackerException):
    message = _("%(message)s")


class FlowClassiferCreationFailed(exceptions.TackerException):
    message = _("%(message)s")


NAME_MAX_LEN = 255

RESOURCE_ATTRIBUTE_MAP = {

    'vims': {
        'id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {'type:uuid': None},
            'is_visible': True,
            'primary_key': True,
        },
        'tenant_id': {
            'allow_post': True,
            'allow_put': False,
            'validate': {'type:string': None},
            'required_by_policy': True,
            'is_visible': True
        },
        'type': {
            'allow_post': True,
            'allow_put': False,
            'validate': {'type:not_empty_string': None},
            'is_visible': True
        },
        'auth_url': {
            'allow_post': True,
            'allow_put': False,
            'validate': {'type:string': None},
            'is_visible': True
        },
        'auth_cred': {
            'allow_post': True,
            'allow_put': True,
            'validate': {'type:dict_not_empty': None},
            'is_visible': True,
        },
        'vim_project': {
            'allow_post': True,
            'allow_put': True,
            'validate': {'type:dict_not_empty': None},
            'is_visible': True,
        },
        'name': {
            'allow_post': True,
            'allow_put': True,
            'validate': {'type:string': NAME_MAX_LEN},
            'is_visible': True,
        },
        'description': {
            'allow_post': True,
            'allow_put': True,
            'validate': {'type:string': None},
            'is_visible': True,
            'default': '',
        },
        'status': {
            'allow_post': False,
            'allow_put': False,
            'validate': {'type:string': None},
            'is_visible': True,
        },
        'placement_attr': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
            'default': None,
        },
        'shared': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': False,
            'convert_to': attr.convert_to_boolean,
            'required_by_policy': True
        },
        'is_default': {
            'allow_post': True,
            'allow_put': True,
            'is_visible': True,
            'validate': {'type:boolean': None},
            'default': False
        },
        'created_at': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
        },
        'updated_at': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
        },
    },

    'vnffgds': {
        'id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {'type:uuid': None},
            'is_visible': True,
            'primary_key': True,
        },
        'tenant_id': {
            'allow_post': True,
            'allow_put': False,
            'validate': {'type:string': None},
            'required_by_policy': True,
            'is_visible': True,
        },
        'name': {
            'allow_post': True,
            'allow_put': True,
            'validate': {'type:string': NAME_MAX_LEN},
            'is_visible': True,
        },
        'description': {
            'allow_post': True,
            'allow_put': True,
            'validate': {'type:string': None},
            'is_visible': True,
            'default': '',
        },
        'template': {
            'allow_post': True,
            'allow_put': False,
            'convert_to': attr.convert_none_to_empty_dict,
            'validate': {'type:dict_or_nodata': None},
            'is_visible': True,
            'default': None,
        },
        'template_source': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
            'default': 'onboarded'
        }
    },

    'vnffgs': {
        'id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {'type:uuid': None},
            'is_visible': True,
            'primary_key': True
        },
        'tenant_id': {
            'allow_post': True,
            'allow_put': False,
            'validate': {'type:string': None},
            'required_by_policy': True,
            'is_visible': True
        },
        'vnffgd_id': {
            'allow_post': True,
            'allow_put': False,
            'validate': {'type:uuid': None},
            'is_visible': True,
            'default': None
        },
        'name': {
            'allow_post': True,
            'allow_put': True,
            'validate': {'type:string': NAME_MAX_LEN},
            'is_visible': True,
        },
        'description': {
            'allow_post': True,
            'allow_put': True,
            'validate': {'type:string': None},
            'is_visible': True,
            'default': '',
        },
        'vnf_mapping': {
            'allow_post': True,
            'allow_put': True,
            'convert_to': attr.convert_none_to_empty_dict,
            'validate': {'type:dict_or_nodata': None},
            'is_visible': True,
            'default': None,
        },
        'attributes': {
            'allow_post': True,
            'allow_put': True,
            'convert_to': attr.convert_none_to_empty_dict,
            'validate': {'type:dict_or_nodata': None},
            'is_visible': True,
            'default': None,
        },
        'symmetrical': {
            'allow_post': True,
            'allow_put': True,
            'is_visible': True,
            'validate': {'type:boolean': None},
            'default': False,
        },
        'forwarding_paths': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
        },
        'status': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
        },
        'vnffgd_template': {
            'allow_post': True,
            'allow_put': True,
            'validate': {'type:dict_or_nodata': None},
            'is_visible': True,
            'default': None,
        },
        'ns_id': {
            'allow_post': True,
            'allow_put': False,
            'is_visible': True,
            'default': None,
        },
    },

    'nfps': {
        'id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {'type:uuid': None},
            'is_visible': True,
            'primary_key': True
        },
        'tenant_id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {'type:string': None},
            'required_by_policy': True,
            'is_visible': True
        },
        'vnffg_id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {'type:uuid': None},
            'is_visible': True,
        },
        'name': {
            'allow_post': False,
            'allow_put': False,
            'validate': {'type:string': None},
            'is_visible': True,
        },
        'classifier_id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {'type:uuid': None},
            'is_visible': True,
        },
        'chain_id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {'type:uuid': None},
            'is_visible': True,
        },
        'path_id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {'type:string': None},
            'is_visible': True,
        },
        'symmetrical': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
            'validate': {'type:boolean': None},
            'default': False,
        },
        'status': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
        },
    },
    'sfcs': {
        'id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {'type:uuid': None},
            'is_visible': True,
            'primary_key': True
        },
        'tenant_id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {'type:string': None},
            'required_by_policy': True,
            'is_visible': True
        },
        'nfp_id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {'type:uuid': None},
            'is_visible': True,
        },
        'instance_id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {'type:uuid': None},
            'is_visible': True,
        },
        'chain': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
        },
        'path_id': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
        },
        'symmetrical': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
            'validate': {'type:boolean': None},
            'default': False,
        },
        'status': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
        },
    },
    'classifiers': {
        'id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {'type:uuid': None},
            'is_visible': True,
            'primary_key': True
        },
        'tenant_id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {'type:string': None},
            'required_by_policy': True,
            'is_visible': True
        },
        'nfp_id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {'type:uuid': None},
            'is_visible': True,
        },
        'instance_id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {'type:uuid': None},
            'is_visible': True,
        },
        'match': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
        },
        'chain_id': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
        },
        'status': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
        },
        'name': {
            'allow_post': True,
            'allow_put': True,
            'validate': {'type:string': NAME_MAX_LEN},
            'is_visible': True,
        },
    },

    'nsds': {
        'id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {'type:uuid': None},
            'is_visible': True,
            'primary_key': True,
        },
        'tenant_id': {
            'allow_post': True,
            'allow_put': False,
            'validate': {'type:string': None},
            'required_by_policy': True,
            'is_visible': True,
        },
        'name': {
            'allow_post': True,
            'allow_put': True,
            'validate': {'type:string': NAME_MAX_LEN},
            'is_visible': True,
        },
        'description': {
            'allow_post': True,
            'allow_put': True,
            'validate': {'type:string': None},
            'is_visible': True,
            'default': '',
        },
        'created_at': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
        },
        'updated_at': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
        },
        'attributes': {
            'allow_post': True,
            'allow_put': False,
            'convert_to': attr.convert_none_to_empty_dict,
            'validate': {'type:dict_or_nodata': None},
            'is_visible': True,
            'default': None,
        },
        'template_source': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
            'default': 'onboarded'
        },

    },

    'nss': {
        'id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {'type:uuid': None},
            'is_visible': True,
            'primary_key': True,
        },
        'tenant_id': {
            'allow_post': True,
            'allow_put': False,
            'validate': {'type:string': None},
            'required_by_policy': True,
            'is_visible': True,
        },
        'name': {
            'allow_post': True,
            'allow_put': True,
            'validate': {'type:string': NAME_MAX_LEN},
            'is_visible': True,
        },
        'description': {
            'allow_post': True,
            'allow_put': True,
            'validate': {'type:string': NAME_MAX_LEN},
            'is_visible': True,
            'default': '',
        },
        'created_at': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
        },
        'updated_at': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
        },
        'vnf_ids': {
            'allow_post': True,
            'allow_put': False,
            'validate': {'type:string': None},
            'is_visible': True,
            'default': '',
        },
        'vnffg_ids': {
            'allow_post': True,
            'allow_put': False,
            'validate': {'type:string': None},
            'is_visible': True,
            'default': '',
        },
        'nsd_id': {
            'allow_post': True,
            'allow_put': False,
            'validate': {'type:uuid': None},
            'is_visible': True,
            'default': None,
        },
        'placement_attr': {
            'allow_post': True,
            'allow_put': False,
            'validate': {'type:dict_or_none': None},
            'is_visible': True,
            'default': {},
        },
        'vim_id': {
            'allow_post': True,
            'allow_put': False,
            'validate': {'type:string': None},
            'is_visible': True,
            'default': '',
        },
        'status': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
        },
        'error_reason': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True,
        },
        'attributes': {
            'allow_post': True,
            'allow_put': False,
            'convert_to': attr.convert_none_to_empty_dict,
            'validate': {'type:dict_or_nodata': None},
            'is_visible': True,
            'default': None,
        },
        'mgmt_ip_addresses': {
            'allow_post': False,
            'allow_put': False,
            'convert_to': attr.convert_none_to_empty_dict,
            'validate': {'type:dict_or_nodata': None},
            'is_visible': True,
        },
        'nsd_template': {
            'allow_post': True,
            'allow_put': False,
            'validate': {'type:dict_or_nodata': None},
            'is_visible': True,
            'default': None,
        },
    },

}


class Nfvo(extensions.ExtensionDescriptor):
    @classmethod
    def get_name(cls):
        return 'NFV Orchestrator'

    @classmethod
    def get_alias(cls):
        return 'NFVO'

    @classmethod
    def get_description(cls):
        return "Extension for NFV Orchestrator"

    @classmethod
    def get_namespace(cls):
        return 'https://wiki.openstack.org/Tacker'

    @classmethod
    def get_updated(cls):
        return "2015-12-21T10:00:00-00:00"

    @classmethod
    def get_resources(cls):
        special_mappings = {}
        plural_mappings = resource_helper.build_plural_mappings(
            special_mappings, RESOURCE_ATTRIBUTE_MAP)
        attr.PLURALS.update(plural_mappings)
        return resource_helper.build_resource_info(
            plural_mappings, RESOURCE_ATTRIBUTE_MAP, constants.NFVO,
            translate_name=True)

    @classmethod
    def get_plugin_interface(cls):
        return NFVOPluginBase

    def update_attributes_map(self, attributes):
        super(Nfvo, self).update_attributes_map(
            attributes, extension_attrs_map=RESOURCE_ATTRIBUTE_MAP)

    def get_extended_resources(self, version):
        version_map = {'1.0': RESOURCE_ATTRIBUTE_MAP}
        return version_map.get(version, {})


class NFVOPluginBase(service_base.NFVPluginBase, metaclass=abc.ABCMeta):
    def get_plugin_name(self):
        return constants.NFVO

    def get_plugin_type(self):
        return constants.NFVO

    def get_plugin_description(self):
        return 'Tacker NFV Orchestrator plugin'

    @abc.abstractmethod
    def create_vim(self, context, vim):
        pass

    @abc.abstractmethod
    def update_vim(self, context, vim_id, vim):
        pass

    @abc.abstractmethod
    def delete_vim(self, context, vim_id):
        pass

    @abc.abstractmethod
    def get_vim(self, context, vim_id, fields=None, mask_password=True):
        pass

    @abc.abstractmethod
    def get_vims(self, context, filters=None, fields=None):
        pass

    def get_vim_by_name(self, context, vim_name, fields=None,
                        mask_password=True):
        raise NotImplementedError()

    def get_default_vim(self, context):
        pass



# Copyright 2018 NTT DATA
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import yaml

from tackerclient.common import exceptions

from tacker.tests import constants
from tacker.tests.functional import base
from tacker.tests.utils import read_file


class VnfTestCreate(base.BaseTackerTest):
    def _test_create_delete_vnf(self, vnf_name, vnfd_name,
                                placement_policy, vdu_name,
                                vnf_expected_status="ACTIVE"):
        input_yaml = read_file(vnfd_name + '.yaml')
        tosca_dict = yaml.safe_load(input_yaml)
        tosca_arg = {'vnfd': {'name': vnfd_name,
                     'attributes': {'vnfd': tosca_dict}}}

        # Create vnfd with tosca template
        vnfd_instance = self.client.create_vnfd(body=tosca_arg)
        self.assertIsNotNone(vnfd_instance)

        # Get vnfd_id
        vnfd_id = vnfd_instance['vnfd']['id']

        # Add vnfd delete to cleanup job so that if vnf_instance fails to
        # create then it will be cleaned-up automatically in tearDown()
        self.addCleanup(self.client.delete_vnfd, vnfd_id)

        vnf_arg = {'vnf': {'vnfd_id': vnfd_id, 'name': vnf_name}}

        # Create vnf instance
        vnf_instance = self.client.create_vnf(body=vnf_arg)
        vnf_id = vnf_instance['vnf']['id']

        # Delete vnf_instance after tearDown
        self.addCleanup(self.wait_until_vnf_delete, vnf_id,
                        constants.VNF_CIRROS_DELETE_TIMEOUT)
        self.addCleanup(self.client.delete_vnf, vnf_id)

        self.validate_vnf_instance(vnfd_instance, vnf_instance)

        self.wait_until_vnf_status(
            vnf_id,
            vnf_expected_status,
            constants.VNF_CIRROS_CREATE_TIMEOUT,
            constants.ACTIVE_SLEEP_TIME)

        # VDU names are random generated names with initials of *vnf_name*.
        # Search the instance_list with *vdu_name* regular expression
        opts = {
            "name": vdu_name
        }
        compute_hosts = []
        vm_statuses = []
        for server in self.novaclient().servers.list(search_opts=opts):
            instance_host = getattr(server,
                                    "OS-EXT-SRV-ATTR:hypervisor_hostname")
            vm_statuses.append(getattr(server, "status"))
            compute_hosts.append(instance_host)

        # check vnf placement policies
        if placement_policy == 'affinity':
            # check "compute_hosts" is not empty
            self.assertTrue(compute_hosts)

            # Get the first compute_host on which VDU is deployed and compare
            # it with other compute hosts of VDU's
            compute_host = compute_hosts[0]
            for vnf in compute_hosts:
                self.assertEqual(compute_host, vnf)
        elif placement_policy == 'anti-affinity':
            if vnf_expected_status == "ERROR":
                # Check one of the VM should be in "ERROR" status
                # and instance host should be None.
                self.assertIn("ERROR", vm_statuses)
                self.assertIn(None, compute_hosts)
            else:
                distinct_comp_hosts = set(compute_hosts)
                self.assertEqual(len(compute_hosts), len(distinct_comp_hosts))

    def test_create_delete_vnf_with_placement_policy_affinity(self):
        self._test_create_delete_vnf(
            vnf_name='test_vnf_with_placement_policy_affinity',
            vnfd_name='sample-tosca-vnfd-placement-policy-affinity',
            vdu_name='affinity-vdu',
            placement_policy='affinity')

    def test_create_delete_vnf_with_placement_policy_anti_affinity(self):
        self._test_create_delete_vnf(
            vnf_name='test_vnf_with_placement_policy_anti_affinity',
            vnfd_name='sample-tosca-vnfd-placement-policy-anti-affinity',
            vdu_name='anti-affinity-vdu-multi-comp-nodes',
            placement_policy='anti-affinity')

    def test_vnf_with_policy_anti_affinity_insufficient_comp_nodes(self):
        self._test_create_delete_vnf(
            vnf_name='test_vnf_anti_affinity_insufficient_comp_nodes',
            vnfd_name='sample-tosca-vnfd-anti-affinity-multi-vdu',
            vdu_name='anti-affinity-vdu-insufficient-comp-nodes',
            placement_policy='anti-affinity',
            vnf_expected_status="ERROR")

    def test_vnf_with_placement_policy_invalid(self):
        exc = self.assertRaises(
            exceptions.BadRequest,
            self._test_create_delete_vnf,
            vnf_name='test_vnf_with_placement_policy_invalid',
            vnfd_name='sample-tosca-vnfd-placement-policy-invalid',
            vdu_name='invalid-placement-policy-vdu',
            placement_policy='invalid')
        self.assertIn('"invalid"', exc.message)
        self.assertIn('The value "invalid" of property "policy" is '
                      'not valid. Expected a value from "[affinity, '
                      'anti-affinity]".', exc.message)



# Copyright 2020 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

# flake8: noqa: E402

"""add_vnflcm_subscription

Revision ID: c47a733f425a
Revises: 8a7ca803e0d0 
Create Date: 2020-08-27 14:18:43.907565

"""

# revision identifiers, used by Alembic.
revision = 'c47a733f425a'
down_revision = '8a7ca803e0d0'

from alembic import op
import sqlalchemy as sa
from sqlalchemy import Boolean

from tacker.db import types


def upgrade(active_plugins=None, options=None):
    op.create_table(
        'vnf_lcm_subscriptions',
        sa.Column('id', types.Uuid(length=36), nullable=False),
        sa.Column('callback_uri', sa.String(length=255), nullable=False),
        sa.Column('subscription_authentication', sa.JSON(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', Boolean, default=False),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )

    noti_str = "json_unquote(json_extract('filter','$.notificationTypes'))"
    sta_str = "json_unquote(json_extract('filter','$.operationStates'))"
    op.create_table(
        'vnf_lcm_filters',
        sa.Column('id', sa.Integer, autoincrement=True, nullable=False),
        sa.Column('subscription_uuid', sa.String(length=36), nullable=False),
        sa.Column('filter', sa.JSON(), nullable=False),
        sa.Column('notification_types',
                  sa.LargeBinary(length=65536),
                  sa.Computed(noti_str)),
        sa.Column('notification_types_len',
                  sa.Integer,
                  sa.Computed("ifnull(json_length('notification_types'),0)")),
        sa.Column('operation_states',
                  sa.LargeBinary(length=65536),
                  sa.Computed(sta_str)),
        sa.Column('operation_states_len',
                  sa.Integer,
                  sa.Computed("ifnull(json_length('operation_states'),0)")),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(['subscription_uuid'],
                                ['vnf_lcm_subscriptions.id'], ),
        mysql_engine='InnoDB'
    )

    op.create_table(
        'vnf_lcm_op_occs',
        sa.Column('id', types.Uuid(length=36), nullable=False),
        sa.Column('operation_state', sa.String(length=255), nullable=False),
        sa.Column('state_entered_time',  sa.DateTime(), nullable=False),
        sa.Column('start_time',  sa.DateTime(), nullable=False),
        sa.Column('vnf_instance_id', types.Uuid(length=36), nullable=False),
        sa.Column('operation', sa.String(length=255), nullable=False),
        sa.Column('is_automatic_invocation', sa.Boolean, nullable=False),
        sa.Column('operation_params', sa.JSON(), nullable=True),
        sa.Column('error', sa.JSON(), nullable=True),
        sa.Column('deleted', Boolean, default=False),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(['vnf_instance_id'],
                                ['vnf_instances.id'], ),
        mysql_engine='InnoDB'
    )



# Copyright (C) 2022 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""Basic Linux commands intented to be used in unittests"""

from oslo_concurrency import processutils

import tacker.privileged


@tacker.privileged.default.entrypoint
def pwd():
    pwd = processutils.execute('pwd')
    return pwd


@tacker.privileged.default.entrypoint
def ls():
    ls = processutils.execute('ls')
    return ls



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from oslo_log import log as logging

from tacker.plugins.common import constants
from tacker.vnfm.policy_actions import abstract_action
from tacker.vnfm import utils as vnfm_utils

LOG = logging.getLogger(__name__)


class VNFActionLog(abstract_action.AbstractPolicyAction):
    def get_type(self):
        return 'log'

    def get_name(self):
        return 'log'

    def get_description(self):
        return 'Tacker VNF logging policy'

    def execute_action(self, plugin, context, vnf_dict, args):
        vnf_id = vnf_dict['id']
        LOG.error('vnf %s dead', vnf_id)
        vnfm_utils.log_events(context, vnf_dict,
                              constants.RES_EVT_MONITOR,
                              "ActionLogOnly invoked")


class VNFActionLogAndKill(abstract_action.AbstractPolicyAction):
    def get_type(self):
        return 'log_and_kill'

    def get_name(self):
        return 'log_and_kill'

    def get_description(self):
        return 'Tacker VNF log_and_kill policy'

    def execute_action(self, plugin, context, vnf_dict, args):
        vnfm_utils.log_events(context, vnf_dict,
                              constants.RES_EVT_MONITOR,
                              "ActionLogAndKill invoked")
        vnf_id = vnf_dict['id']
        if plugin._mark_vnf_dead(vnf_dict['id']):
            if vnf_dict['attributes'].get('monitoring_policy'):
                plugin._vnf_monitor.mark_dead(vnf_dict['id'])
            plugin.delete_vnf(context, vnf_id)
        LOG.error('vnf %s dead', vnf_id)



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker.sol_refactored.objects import fields


# NFV-SOL 003
# - v2.6.1 9.5.4.3 (API version: 1.3.0)
# - v2.7.1 9.5.4.3 (API version: 1.3.0)
# - v2.8.1 9.5.4.3 (API version: 1.3.0)
# - v3.3.1 9.5.4.3 (API version: 1.4.0)
class GrantedLcmOperationType(fields.BaseTackerEnum):
    INSTANTIATE = 'INSTANTIATE'
    SCALE = 'SCALE'
    SCALE_TO_LEVEL = 'SCALE_TO_LEVEL'
    CHANGE_FLAVOUR = 'CHANGE_FLAVOUR'
    TERMINATE = 'TERMINATE'
    HEAL = 'HEAL'
    OPERATE = 'OPERATE'
    CHANGE_EXT_CONN = 'CHANGE_EXT_CONN'
    CHANGE_VNFPKG = 'CHANGE_VNFPKG'  # since 1.4.0
    CREATE_SNAPSHOT = 'CREATE_SNAPSHOT'  # since 1.4.0
    REVERT_TO_SNAPSHOT = 'REVERT_TO_SNAPSHOT'  # since 1.4.0

    ALL = (INSTANTIATE, SCALE, SCALE_TO_LEVEL, CHANGE_FLAVOUR,
           TERMINATE, HEAL, OPERATE, CHANGE_EXT_CONN, CHANGE_VNFPKG,
           CREATE_SNAPSHOT, REVERT_TO_SNAPSHOT)


class GrantedLcmOperationTypeField(fields.BaseEnumField):
    AUTO_TYPE = GrantedLcmOperationType()



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import os
import sys

from tacker.sol_refactored.common import http_client


# NOTE: This is the tool which intended to use it about APIs under development.
# This is not maintained at all, but reuse the code for the new APIs
# development.

auth_url = "http://127.0.0.1/identity/v3"
username = "nfv_user"
password = "devstack"
project_name = "nfv"
domain_name = "Default"
user_domain_name = "Default"
project_domain_name = "Default"


class Client(object):

    def __init__(self, path):
        self.path = path
        self.auth = http_client.KeystonePasswordAuthHandle(
            auth_url=auth_url,
            username=username,
            password=password,
            project_name=project_name,
            user_domain_name=user_domain_name,
            project_domain_name=project_domain_name)

        self.client = http_client.HttpClient(self.auth)

    def print(self, resp, body):
        print(resp.status_code)
        print(resp.headers)
        print()
        print(body)

    def create(self, req_body):
        resp, body = self.client.do_request(
            self.path, "POST", body=req_body, version="2.0.0")
        self.print(resp, body)

    def list(self, req_body):
        if req_body is not None:
            resp, body = self.client.do_request(
                self.path, "GET", version="2.0.0", params=req_body)
        else:
            resp, body = self.client.do_request(
                self.path, "GET", version="2.0.0")
        self.print(resp, body)

    def show(self, id):
        resp, body = self.client.do_request(
            self.path + '/' + id, "GET", version="2.0.0")
        self.print(resp, body)

    def delete(self, id):
        resp, body = self.client.do_request(
            self.path + '/' + id, "DELETE", version="2.0.0")
        self.print(resp, body)

    def update(self, id, req_body):
        resp, body = self.client.do_request(
            self.path + '/' + id, "PATCH", body=req_body, version="2.0.0")
        self.print(resp, body)

    def inst(self, id, req_body):
        path = self.path + '/' + id + '/instantiate'
        resp, body = self.client.do_request(
            path, "POST", body=req_body, version="2.0.0")
        self.print(resp, body)

    def term(self, id, req_body):
        path = self.path + '/' + id + '/terminate'
        resp, body = self.client.do_request(
            path, "POST", body=req_body, version="2.0.0")
        self.print(resp, body)

    def scale(self, id, req_body):
        path = self.path + '/' + id + '/scale'
        resp, body = self.client.do_request(
            path, "POST", body=req_body, version="2.0.0")
        self.print(resp, body)

    def heal(self, id, req_body):
        path = self.path + '/' + id + '/heal'
        resp, body = self.client.do_request(
            path, "POST", body=req_body, version="2.0.0")
        self.print(resp, body)

    def chg_ext_conn(self, id, req_body):
        path = self.path + '/' + id + '/change_ext_conn'
        resp, body = self.client.do_request(
            path, "POST", body=req_body, version="2.0.0")
        self.print(resp, body)

    def retry(self, id):
        path = self.path + '/' + id + '/retry'
        resp, body = self.client.do_request(path, "POST", version="2.0.0")
        self.print(resp, body)

    def rollback(self, id):
        path = self.path + '/' + id + '/rollback'
        resp, body = self.client.do_request(path, "POST", version="2.0.0")
        self.print(resp, body)

    def fail(self, id):
        path = self.path + '/' + id + '/fail'
        resp, body = self.client.do_request(path, "POST", version="2.0.0")
        self.print(resp, body)


def usage():
    print("usage: cli resource action [arg...]")
    print("  inst create body(path of content)")
    print("  inst list [body(path of content)]")
    print("  inst show {id}")
    print("  inst delete {id}")
    print("  inst update {id} body(path of content)")
    print("  inst inst {id} body(path of content)")
    print("  inst term {id} body(path of content)")
    print("  inst scale {id} body(path of content)")
    print("  inst heal {id} body(path of content)")
    print("  inst chg_ext_conn {id} body(path of content)")
    print("  subsc create body(path of content)")
    print("  subsc list [body(path of content)]")
    print("  subsc show {id}")
    print("  subsc delete {id}")
    print("  lcmocc list [body(path of content)]")
    print("  lcmocc show {id}")
    print("  lcmocc delete {id}")
    print("  lcmocc retry {id}")
    print("  lcmocc rollback {id}")
    print("  lcmocc fail {id}")
    os._exit(1)


def get_body(arg):
    with open(arg) as fp:
        return json.load(fp)


if __name__ == '__main__':
    if len(sys.argv) < 3:
        usage()
    resource = sys.argv[1]
    action = sys.argv[2]

    if resource == "inst":
        if action not in ["create", "list", "show", "delete", "update",
                          "inst", "term", "scale", "heal", "chg_ext_conn"]:
            usage()
        client = Client("/vnflcm/v2/vnf_instances")
    elif resource == "subsc":
        if action not in ["create", "list", "show", "delete"]:
            usage()
        client = Client("/vnflcm/v2/subscriptions")
    elif resource == "lcmocc":
        if action not in ["list", "show", "delete", "retry", "rollback",
                          "fail"]:
            usage()
        client = Client("/vnflcm/v2/vnf_lcm_op_occs")
    else:
        usage()

    if action == "create":
        if len(sys.argv) != 4:
            usage()
        client.create(get_body(sys.argv[3]))
    elif action == "list":
        body = None
        if len(sys.argv) == 4:
            body = get_body(sys.argv[3])
        elif len(sys.argv) != 3:
            usage()
        client.list(body)
    elif action == "show":
        if len(sys.argv) != 4:
            usage()
        client.show(sys.argv[3])
    elif action == "delete":
        if len(sys.argv) != 4:
            usage()
        client.delete(sys.argv[3])
    elif action == "update":
        if len(sys.argv) != 5:
            usage()
        client.update(sys.argv[3], get_body(sys.argv[4]))
    elif action == "inst":
        if len(sys.argv) != 5:
            usage()
        client.inst(sys.argv[3], get_body(sys.argv[4]))
    elif action == "term":
        if len(sys.argv) != 5:
            usage()
        client.term(sys.argv[3], get_body(sys.argv[4]))
    elif action == "scale":
        if len(sys.argv) != 5:
            usage()
        client.scale(sys.argv[3], get_body(sys.argv[4]))
    elif action == "heal":
        if len(sys.argv) != 5:
            usage()
        client.heal(sys.argv[3], get_body(sys.argv[4]))
    elif action == "chg_ext_conn":
        if len(sys.argv) != 5:
            usage()
        client.chg_ext_conn(sys.argv[3], get_body(sys.argv[4]))
    elif action == "retry":
        if len(sys.argv) != 4:
            usage()
        client.retry(sys.argv[3])
    elif action == "rollback":
        if len(sys.argv) != 4:
            usage()
        client.rollback(sys.argv[3])
    elif action == "fail":
        if len(sys.argv) != 4:
            usage()
        client.fail(sys.argv[3])



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_policy import policy


POLICY_NAME = 'os_nfv_orchestration_api_v2:vnf_instances:{}'
RULE_ANY = '@'

V2_PATH = '/vnflcm/v2'
API_VERSIONS_PATH = V2_PATH + '/api_versions'
VNF_INSTANCES_PATH = V2_PATH + '/vnf_instances'
VNF_INSTANCES_ID_PATH = VNF_INSTANCES_PATH + '/{vnfInstanceId}'
SUBSCRIPTIONS_PATH = V2_PATH + '/subscriptions'
SUBSCRIPTIONS_ID_PATH = VNF_INSTANCES_PATH + '/{subscriptionId}'
VNF_LCM_OP_OCCS_PATH = V2_PATH + '/vnf_lcm_op_occs'
VNF_LCM_OP_OCCS_ID_PATH = VNF_LCM_OP_OCCS_PATH + '/{vnfLcmOpOccId}'

rules = [
    policy.DocumentedRuleDefault(
        name=POLICY_NAME.format('api_versions'),
        check_str=RULE_ANY,
        description="Get API Versions.",
        operations=[
            {'method': 'GET',
             'path': API_VERSIONS_PATH}
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_NAME.format('create'),
        check_str=RULE_ANY,
        description="Creates vnf instance.",
        operations=[
            {'method': 'POST',
             'path': VNF_INSTANCES_PATH}
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_NAME.format('index'),
        check_str=RULE_ANY,
        description="Query VNF instances.",
        operations=[
            {'method': 'GET',
             'path': VNF_INSTANCES_PATH}
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_NAME.format('show'),
        check_str=RULE_ANY,
        description="Query an Individual VNF instance.",
        operations=[
            {'method': 'GET',
             'path': VNF_INSTANCES_ID_PATH}
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_NAME.format('delete'),
        check_str=RULE_ANY,
        description="Delete an Individual VNF instance.",
        operations=[
            {'method': 'DELETE',
             'path': VNF_INSTANCES_ID_PATH}
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_NAME.format('update'),
        check_str=RULE_ANY,
        description="Modify vnf instance information.",
        operations=[
            {'method': 'PATCH',
             'path': VNF_INSTANCES_ID_PATH}
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_NAME.format('instantiate'),
        check_str=RULE_ANY,
        description="Instantiate vnf instance.",
        operations=[
            {'method': 'POST',
             'path': VNF_INSTANCES_ID_PATH + '/instantiate'}
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_NAME.format('terminate'),
        check_str=RULE_ANY,
        description="Terminate vnf instance.",
        operations=[
            {'method': 'POST',
             'path': VNF_INSTANCES_ID_PATH + '/terminate'}
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_NAME.format('scale'),
        check_str=RULE_ANY,
        description="Scale vnf instance.",
        operations=[
            {'method': 'POST',
             'path': VNF_INSTANCES_ID_PATH + '/scale'}
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_NAME.format('heal'),
        check_str=RULE_ANY,
        description="Heal vnf instance.",
        operations=[
            {'method': 'POST',
             'path': VNF_INSTANCES_ID_PATH + '/heal'}
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_NAME.format('change_ext_conn'),
        check_str=RULE_ANY,
        description="Change external vnf connectivity.",
        operations=[
            {'method': 'POST',
             'path': VNF_INSTANCES_ID_PATH + '/change_ext_conn'}
        ]
    ),
    # NOTE: add when the operation supported
    policy.DocumentedRuleDefault(
        name=POLICY_NAME.format('subscription_create'),
        check_str=RULE_ANY,
        description="Create subscription.",
        operations=[
            {'method': 'POST',
             'path': SUBSCRIPTIONS_PATH}
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_NAME.format('subscription_list'),
        check_str=RULE_ANY,
        description="List subscription.",
        operations=[
            {'method': 'GET',
             'path': SUBSCRIPTIONS_PATH}
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_NAME.format('subscription_show'),
        check_str=RULE_ANY,
        description="Show subscription.",
        operations=[
            {'method': 'GET',
             'path': SUBSCRIPTIONS_ID_PATH}
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_NAME.format('subscription_delete'),
        check_str=RULE_ANY,
        description="Delete subscription.",
        operations=[
            {'method': 'DELETE',
             'path': SUBSCRIPTIONS_ID_PATH}
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_NAME.format('lcm_op_occ_list'),
        check_str=RULE_ANY,
        description="List VnfLcmOpOcc.",
        operations=[
            {'method': 'GET',
             'path': VNF_LCM_OP_OCCS_PATH}
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_NAME.format('lcm_op_occ_show'),
        check_str=RULE_ANY,
        description="Show VnfLcmOpOcc.",
        operations=[
            {'method': 'GET',
             'path': VNF_LCM_OP_OCCS_ID_PATH}
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_NAME.format('lcm_op_occ_retry'),
        check_str=RULE_ANY,
        description="Retry VnfLcmOpOcc.",
        operations=[
            {'method': 'POST',
             'path': VNF_LCM_OP_OCCS_ID_PATH + '/retry'}
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_NAME.format('lcm_op_occ_rollback'),
        check_str=RULE_ANY,
        description="Rollback VnfLcmOpOcc.",
        operations=[
            {'method': 'POST',
             'path': VNF_LCM_OP_OCCS_ID_PATH + '/rollback'}
        ]
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_NAME.format('lcm_op_occ_fail'),
        check_str=RULE_ANY,
        description="Fail VnfLcmOpOcc.",
        operations=[
            {'method': 'POST',
             'path': VNF_LCM_OP_OCCS_ID_PATH + '/fail'}
        ]
    ),
    # NOTE: 'DELETE' is not defined in the specification. It is for test
    # use since it is convenient to be able to delete under development.
    # It is available when config parameter
    # v2_vnfm.test_enable_lcm_op_occ_delete set to True.
    policy.DocumentedRuleDefault(
        name=POLICY_NAME.format('lcm_op_occ_delete'),
        check_str=RULE_ANY,
        description="Delete VnfLcmOpOcc.",
        operations=[
            {'method': 'DELETE',
             'path': VNF_LCM_OP_OCCS_ID_PATH}
        ]
    ),
]


def list_rules():
    return rules
